/*
 *  Copyright (C) 2024 Egor Ignatov <egori@altlinux.org>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdarg.h>

#define DOS_MAGIC (uint16_t)(0x5a4d)
#define PE_MAGIC (uint32_t)(0x00004550)

struct DOS_HDR {
  uint16_t magic;
  uint8_t _pad[58];
  uint32_t lfanew;
};

struct SEC_HDR {
  uint8_t name[8];
  uint8_t _pad1[8];
  uint32_t size;
  uint32_t offset;
  uint8_t _pad2[16];
};

struct PE_HDR {
  uint32_t signature;
  struct FILE_HDR {
    uint16_t machine;
    uint16_t number_of_sections;
    uint32_t time_date_stamp;
    uint32_t pointer_to_symbol_table;
    uint32_t number_of_symbols;
    uint16_t size_of_optional_header;
    uint16_t characteristics;
  } file_hdr;
  union {
    struct OPT_HDR32 {
      uint16_t magic;
      uint8_t _pad[222];
    } opt_hdr32;
    struct OPT_HDR64 {
      uint16_t magic;
      uint8_t _pad[238];
    } opt_hdr64;
  };
};

static void fatal(const char *fmt, ...)
{
  va_list args;
  va_start(args, fmt);
  fprintf(stderr, "FATAL: ");
  vfprintf(stderr, fmt, args);
  fprintf(stderr, "\n");
  va_end(args);
  exit(1);
}

static void *xmalloc(size_t num_bytes)
{
    void *ptr = malloc(num_bytes);
    if (!ptr) {
        fatal("xmalloc failed");
    }
    return ptr;
}

static char *read_file(size_t *buf_size, const char *path)
{
  FILE *file = fopen(path, "rb");
  if (!file) {
    return NULL;
  }
  fseek(file, 0, SEEK_END);
  long len = ftell(file);
  if (len < 0) {
    fclose(file);
    return NULL;
  }
  fseek(file, 0, SEEK_SET);
  char *buf = xmalloc(len + 1);
  if (len && fread(buf, len, 1, file) != 1) {
    fclose(file);
    free(buf);
    return NULL;
  }
  fclose(file);
  buf[len] = 0;
  *buf_size = len;
  return buf;
}

int main(int argc, char **argv)
{
  if (argc != 2) {
    fprintf(stderr, "Usage: %s EFI_FILE\n", argv[0]);
    return 1;
  }

  if (strncmp(argv[1], "-h", 2) == 0 || strncmp(argv[1], "--help", 6) == 0) {
    printf("Usage: %s EFI_FILE\n", argv[0]);
    return 0;
  }

  size_t buf_size;
  char *buf = read_file(&buf_size, argv[1]);
  char *buf_end = buf + buf_size;

  if (!buf) {
    fatal("Failed to read file '%s'", argv[1]);
  }

  struct DOS_HDR *dos_hdr = (struct DOS_HDR *)buf;
  if ((char *)&dos_hdr[1] > buf_end) {
    fatal("DOS Header OoB");
  }
  if (dos_hdr->magic != DOS_MAGIC) {
    fatal("DOS Header magic mismatch");
  }

  struct PE_HDR *pe_hdr = (struct PE_HDR *)(buf + dos_hdr->lfanew);
  if ((char *)pe_hdr < buf || (char *)&pe_hdr[1] > buf_end) {
    fatal("PE Header OoB");
  }
  if (pe_hdr->signature != PE_MAGIC) {
    fatal("PE Header magic mismatch");
  }

  struct SEC_HDR *sections = NULL;

  if (pe_hdr->opt_hdr32.magic == 0x10b) {
    sections = (struct SEC_HDR *) &((&pe_hdr->opt_hdr32)[1]);
  } else if (pe_hdr->opt_hdr32.magic == 0x20b) {
    sections = (struct SEC_HDR *) &((&pe_hdr->opt_hdr64)[1]);
  } else {
    fatal("Optional Header magic mismatch");
  }

  if ((char *)sections < buf
      || (char *)&sections[pe_hdr->file_hdr.number_of_sections+1] > buf_end) {
    fatal("Section Headers OoB");
  }

  for (int i = 0; i < pe_hdr->file_hdr.number_of_sections; ++i) {
    if ((buf + sections[i].offset) < buf
        || (buf + sections[i].offset + sections[i].size) > buf_end) {
      fatal("Section %d OoB", i);
    }

    if (memcmp(sections[i].name, ".sbat\0\0\0", 8) == 0) {
      fwrite(buf + sections[i].offset, 1, sections[i].size, stdout);
      return 0;
    }
  }

  return 1;
}
